# -*- coding: utf-8 -*-
"""
Numpy specific extensions
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np


def boolmask(indices, shape=None):
    """
    Constructs an array of booleans where an item is True if its position is in
    ``indices`` otherwise it is False. This can be viewed as the inverse of
    :func:`numpy.where`.

    Args:
        indices (ndarray): list of integer indices

        shape (int | tuple): length of the returned list. If not specified
            the minimal possible shape to incoporate all the indices is used.
            In general, it is best practice to always specify this argument.

    Returns:
        ndarray[int]: mask: mask[idx] is True if idx in indices

    Example:
        >>> indices = [0, 1, 4]
        >>> mask = boolmask(indices, shape=6)
        >>> assert np.all(mask == [True, True, False, False, True, False])
        >>> mask = boolmask(indices)
        >>> assert np.all(mask == [True, True, False, False, True])

    Example:
        >>> indices = np.array([(0, 0), (1, 1), (2, 1)])
        >>> shape = (3, 3)
        >>> mask = boolmask(indices, shape)
        >>> import ubelt as ub  # NOQA
        >>> result = ub.repr2(mask)
        >>> print(result)
        np.array([[ True, False, False],
                  [False,  True, False],
                  [False,  True, False]], dtype=np.bool)
    """
    indices = np.asanyarray(indices)
    if indices.dtype.kind not in {'i', 'u'}:
        indices = indices.astype(np.int)
    if shape is None:
        shape = indices.max() + 1
    mask = np.zeros(shape, dtype=np.bool)
    if mask.ndim > 1:
        mask[tuple(indices.T)] = True
    else:
        mask[indices] = True
    return mask


def iter_reduce_ufunc(ufunc, arrs, out=None, default=None):
    """
    constant memory iteration and reduction

    applys ufunc from left to right over the input arrays

    Args:
        ufunc (Callable): called on each pair of consecutive ndarrays
        arrs (Iterator[ndarray]): iterator of ndarrays
        default (object): return value when iterator is empty

    Returns:
        ndarray:
            if len(arrs) == 0, returns ``default``
            if len(arrs) == 1, returns arrs[0],
            if len(arrs) >= 2, returns
                ufunc(...ufunc(ufunc(arrs[0], arrs[1]), arrs[2]),...arrs[n-1])

    Example:
        >>> arr_list = [
        ...     np.array([0, 1, 2, 3, 8, 9]),
        ...     np.array([4, 1, 2, 3, 4, 5]),
        ...     np.array([0, 5, 2, 3, 4, 5]),
        ...     np.array([1, 1, 6, 3, 4, 5]),
        ...     np.array([0, 1, 2, 7, 4, 5])
        ... ]
        >>> memory = np.array([9, 9, 9, 9, 9, 9])
        >>> gen_memory = memory.copy()
        >>> def arr_gen(arr_list, gen_memory):
        ...     for arr in arr_list:
        ...         gen_memory[:] = arr
        ...         yield gen_memory
        >>> print('memory = %r' % (memory,))
        >>> print('gen_memory = %r' % (gen_memory,))
        >>> ufunc = np.maximum
        >>> res1 = iter_reduce_ufunc(ufunc, iter(arr_list), out=None)
        >>> res2 = iter_reduce_ufunc(ufunc, iter(arr_list), out=memory)
        >>> res3 = iter_reduce_ufunc(ufunc, arr_gen(arr_list, gen_memory), out=memory)
        >>> print('res1       = %r' % (res1,))
        >>> print('res2       = %r' % (res2,))
        >>> print('res3       = %r' % (res3,))
        >>> print('memory     = %r' % (memory,))
        >>> print('gen_memory = %r' % (gen_memory,))
        >>> assert np.all(res1 == res2)
        >>> assert np.all(res2 == res3)
    """
    # Get first item in iterator
    try:
        initial = next(arrs)
    except StopIteration:
        return default
    # Populate the outvariable if specified otherwise make a copy of the first
    # item to be the output memory
    if out is not None:
        out[:] = initial
    else:
        out = initial.copy()
    # Iterate and reduce
    for arr in arrs:
        ufunc(out, arr, out=out)
    return out


def isect_flags(arr, other):
    """
    Check which items in an array intersect with another set of items

    Args:
        arr (ndarray): items to check
        other (Iterable): items to check if they exist in arr

    Returns:
        ndarray: booleans corresponding to arr indicating if that item is
            also contained in other.

    Example:
        >>> arr = np.array([
        >>>     [1, 2, 3, 4],
        >>>     [5, 6, 3, 4],
        >>>     [1, 1, 3, 4],
        >>> ])
        >>> other = np.array([1, 4, 6])
        >>> mask = isect_flags(arr, other)
        >>> print(mask)
        [[ True False False  True]
         [False  True False  True]
         [ True  True False  True]]
    """
    flags = iter_reduce_ufunc(np.logical_or, (arr == item for item in other))
    if flags is None:
        flags = np.zeros(arr.size, dtype=np.bool)
    return flags


def atleast_nd(arr, n, front=False):
    r"""
    View inputs as arrays with at least n dimensions.

    Args:
        arr (array_like):
            An array-like object.  Non-array inputs are converted to arrays.
            Arrays that already have n or more dimensions are preserved.

        n (int):
            number of dimensions to ensure

        front (bool, default=False):
            if True new dimensions are added to the front of the array.
            otherwise they are added to the back.

    Returns:
        ndarray :
            An array with ``a.ndim >= n``.  Copies are avoided where possible,
            and views with three or more dimensions are returned.  For example,
            a 1-D array of shape ``(N,)`` becomes a view of shape
            ``(1, N, 1)``, and a 2-D array of shape ``(M, N)`` becomes a view
            of shape ``(M, N, 1)``.

    See Also:
        numpy.atleast_1d, numpy.atleast_2d, numpy.atleast_3d

    Example:
        >>> n = 2
        >>> arr = np.array([1, 1, 1])
        >>> arr_ = atleast_nd(arr, n)
        >>> import ubelt as ub  # NOQA
        >>> result = ub.repr2(arr_.tolist(), nl=0)
        >>> print(result)
        [[1], [1], [1]]

    Example:
        >>> n = 4
        >>> arr1 = [1, 1, 1]
        >>> arr2 = np.array(0)
        >>> arr3 = np.array([[[[[1]]]]])
        >>> arr1_ = atleast_nd(arr1, n)
        >>> arr2_ = atleast_nd(arr2, n)
        >>> arr3_ = atleast_nd(arr3, n)
        >>> import ubelt as ub  # NOQA
        >>> result1 = ub.repr2(arr1_.tolist(), nl=0)
        >>> result2 = ub.repr2(arr2_.tolist(), nl=0)
        >>> result3 = ub.repr2(arr3_.tolist(), nl=0)
        >>> result = '\n'.join([result1, result2, result3])
        >>> print(result)
        [[[[1]]], [[[1]]], [[[1]]]]
        [[[[0]]]]
        [[[[[1]]]]]

    Ignore:
        # Hmm, mine is actually faster
        %timeit atleast_nd(arr, 3)
        %timeit np.atleast_3d(arr)

    Benchmark:

        import ubelt
        N = 100

        t1 = ubelt.Timerit(N, label='mine')
        for timer in t1:
            arr = np.empty((10, 10))
            with timer:
                atleast_nd(arr, 3)

        t2 = ubelt.Timerit(N, label='baseline')
        for timer in t2:
            arr = np.empty((10, 10))
            with timer:
                np.atleast_3d(arr)
    """
    arr_ = np.asanyarray(arr)
    ndims = len(arr_.shape)
    if n is not None and ndims <  n:
        # append the required number of dimensions to the front or back
        if front:
            expander = (None,) * (n - ndims) + (Ellipsis,)
        else:
            expander = (Ellipsis,) + (None,) * (n - ndims)
        arr_ = arr_[expander]
    return arr_


def argmaxima(arr, num, axis=None, ordered=True):
    """
    Returns the top ``num`` maximum indicies.

    This can be significantly faster than using argsort.

    Args:
        arr (ndarray): input array
        num (int): number of maximum indices to return
        axis (int|None): axis to find maxima over. If None this is equivalent
            to using arr.ravel().
        ordered (bool): if False, returns the maximum elements in an arbitrary
            order, otherwise they are in decending order. (Setting this to
            false is a bit faster).

    TODO:
        - [ ] if num is None, return arg for all values equal to the maximum

    Returns:
        ndarray

    Example:
        >>> # Test cases with axis=None
        >>> arr = (np.random.rand(100) * 100).astype(np.int)
        >>> for num in range(0, len(arr) + 1):
        >>>     idxs = argmaxima(arr, num)
        >>>     idxs2 = argmaxima(arr, num, ordered=False)
        >>>     assert np.all(arr[idxs] == np.array(sorted(arr)[::-1][:len(idxs)])), 'ordered=True must return in order'
        >>>     assert sorted(idxs2) == sorted(idxs), 'ordered=False must return the right idxs, but in any order'

    Example:
        >>> # Test cases with axis
        >>> arr = (np.random.rand(3, 5, 7) * 100).astype(np.int)
        >>> for axis in range(len(arr.shape)):
        >>>     for num in range(0, len(arr) + 1):
        >>>         idxs = argmaxima(arr, num, axis=axis)
        >>>         idxs2 = argmaxima(arr, num, ordered=False, axis=axis)
        >>>         assert idxs.shape[axis] == num
        >>>         assert idxs2.shape[axis] == num
    """
    # if axis is not None:
    #     raise NotImplementedError('axis must be None for now')
    # Gets top N maximum or minimum indices
    if num < 0:
        raise IndexError
    if axis is None:
        axis_size = arr.size
        if num == 0:
            return np.empty(0, dtype=np.int)
        elif num == 1:
            # very fast
            idxs = np.array([arr.argmax(axis=axis)])
        elif num < axis_size:
            # argpartition is almost what we want, and its faster than argsort
            kth = axis_size - num
            part_idxs = np.argpartition(arr, kth=kth, axis=axis)
            idxs = part_idxs[kth:]
            if ordered:
                sortx = arr.take(idxs, axis=axis).argsort()[::-1]
                idxs = idxs.take(sortx)
        else:
            # sort all the indices
            if ordered:
                idxs = arr.argsort(axis=axis)
                idxs = idxs[::-1][0:num]
            else:
                # Arbitrary order is allowed, so cheat
                idxs = np.arange(arr.size)
    else:
        axis_size = arr.shape[axis]
        if num == 0:
            new_shape = list(arr.shape)
            new_shape[axis] = 0
            return np.empty(new_shape, dtype=np.int)
        elif num == 1:
            newshape = list(arr.shape)
            newshape[axis] = 1
            idxs = arr.argmax(axis=axis).reshape(*newshape)
        elif num < axis_size:
            # TODO: is there a better implementation for this case?
            kth = axis_size - num
            part_idxs = np.argpartition(arr, kth=kth, axis=axis)
            fancy_index = [slice(None)] * (axis + 1)
            fancy_index[axis] = slice(kth, None)
            idxs = part_idxs[tuple(fancy_index)]
            if ordered:
                # move the axis of interest to the back
                idxs_swap = idxs.swapaxes(-1, axis)
                idxs2d = idxs_swap.reshape(-1, num)
                arr2d = arr.swapaxes(-1, axis).reshape(-1, arr.shape[axis])
                # now ensure each row is in order
                new_idxs2d = []
                for a, i in zip(arr2d, idxs2d):
                    sortx = a[i].argsort()[::-1]
                    new_idxs2d.append(i[sortx])
                new_idxs2d = np.array(new_idxs2d)
                # transform back to original shape
                idxs = new_idxs2d.reshape(*idxs_swap.shape).swapaxes(-1, axis)
        else:
            # sort all the indices
            if ordered:
                idxs = arr.argsort(axis=axis)
                fancy_index = [slice(None)] * (axis + 1)
                fancy_index[axis] = slice(None, None, -1)
                idxs = idxs[tuple(fancy_index)]
            else:
                # Arbitrary order is allowed, so cheat
                idxs = np.arange(arr.shape[axis])
                newshape = [1] * len(arr.shape)
                newshape[axis] = arr.shape[axis]
                repeats = list(arr.shape)
                repeats[axis] = 1
                idxs = np.tile(idxs.reshape(*newshape), repeats)
    return idxs


def argminima(arr, num, axis=None, ordered=True):
    """
    Returns the top ``num`` minimum indicies.

    This can be significantly faster than using argsort.

    Args:
        arr (ndarray): input array
        num (int): number of minimum indices to return
        axis (int|None): must be None for now
        ordered (bool): if False, returns the minimum elements in an arbitrary
            order, otherwise they are in ascending order. (Setting this to
            false is a bit faster).

    Example:
        >>> arr = (np.random.rand(100) * 100).astype(np.int)
        >>> for num in range(0, len(arr) + 1):
        >>>     idxs = argminima(arr, num)
        >>>     assert np.all(arr[idxs] == np.array(sorted(arr)[:len(idxs)])), 'ordered=True must return in order'
        >>>     idxs2 = argminima(arr, num, ordered=False)
        >>>     assert sorted(idxs2) == sorted(idxs), 'ordered=False must return the right idxs, but in any order'

    Example:
        >>> arr = (np.random.rand(32, 32) * 100).astype(np.int)
        >>> argminima(arr, 10)
    """
    if axis is not None:
        raise NotImplementedError('axis must be None for now')
    # Gets top N maximum or minimum indices
    if num < 0:
        raise IndexError
    elif num == 0:
        return np.empty(0, dtype=np.int)
    elif num == 1:
        # very fast
        idxs = np.array([arr.argmin(axis=axis)])
    elif num < len(arr):
        # argpartition is almost what we want, and its faster than argsort
        kth = num
        part_idxs = np.argpartition(arr, kth=kth, axis=axis)
        idxs = part_idxs[:kth]
        if ordered:
            sortx = arr.take(idxs, axis=axis).argsort()
            idxs = idxs.take(sortx)
    else:
        # sort all the indices
        if ordered:
            idxs = arr.argsort(axis=axis)[0:num]
        else:
            # Arbitrary order is allowed, so cheat
            idxs = np.arange(arr.size)
    return idxs


def arglexmax(keys, multi=False):
    """
    Find the index of the maximum element in a sequence of keys.

    Args:
        keys (tuple): a k-tuple of k N-dimensional arrays.
            Like np.lexsort the last key in the sequence is used for the
            primary sort order, the second-to-last key for the secondary sort
            order, and so on.

        multi (bool): if True, returns all indices that share the max value

    Returns:
        int | ndarray[int] : either the index or list of indices

    Example:
        >>> k, N = 100, 100
        >>> rng = np.random.RandomState(0)
        >>> keys = [(rng.rand(N) * N).astype(np.int) for _ in range(k)]
        >>> multi_idx = arglexmax(keys, multi=True)
        >>> idxs = np.lexsort(keys)
        >>> assert sorted(idxs[::-1][:len(multi_idx)]) == sorted(multi_idx)

    Benchark:
        >>> import ubelt as ub
        >>> k, N = 100, 100
        >>> rng = np.random
        >>> keys = [(rng.rand(N) * N).astype(np.int) for _ in range(k)]
        >>> for timer in ub.Timerit(100, bestof=10, label='arglexmax'):
        >>>     with timer:
        >>>         arglexmax(keys)
        >>> for timer in ub.Timerit(100, bestof=10, label='lexsort'):
        >>>     with timer:
        >>>         np.lexsort(keys)[-1]
    """
    # Handle keys in reverse order to be consistent with np.lexsort
    reverse_keys = keys[::-1]
    arr = reverse_keys[0]
    breakers = reverse_keys[1:]
    # Look for the maximum value in the first array, and continue using new
    # arrays until a unique maximum index is found.
    _cand_idxs = np.where(arr == arr.max())[0]
    if len(_cand_idxs) > 1:
        for breaker in breakers:
            vals = breaker[_cand_idxs]
            _cand_idxs = _cand_idxs[vals == vals.max()]
            if len(_cand_idxs) == 1:
                break
    # If multiple maximum values are found then either
    # return them all or return an arbitrary one.
    return _cand_idxs if multi else _cand_idxs[0]


def normalize(arr, mode='linear', alpha=None, beta=None, out=None):
    """
    Rebalance signal values via contrast stretching.

    By default linearly stretches array values to minimum and maximum values.

    Args:
        arr (ndarray): array to normalize, usually an image

        out (ndarray | None): output array. Note, that we will create an
            internal floating point copy for integer computations.

        mode (str): either linear or sigmoid.

        alpha (float): Only used if mode=sigmoid.  Division factor
            (pre-sigmoid). If unspecified computed as:
            ``max(abs(old_min - beta), abs(old_max - beta)) / 6.212606``.
            Note this parameter is sensitive to if the input is a float or
            uint8 image.

        beta (float): subtractive factor (pre-sigmoid). This should be the
            intensity of the most interesting bits of the image, i.e. bring
            them to the center (0) of the distribution.
            Defaults to ``(max - min) / 2``.  Note this parameter is sensitive
            to if the input is a float or uint8 image.

    References:
        https://en.wikipedia.org/wiki/Normalization_(image_processing)

    Example:
        >>> raw_f = np.random.rand(8, 8)
        >>> norm_f = normalize(raw_f)

        >>> raw_f = np.random.rand(8, 8) * 100
        >>> norm_f = normalize(raw_f)
        >>> assert np.isclose(norm_f.min(), 0)
        >>> assert np.isclose(norm_f.max(), 1)

        >>> raw_u = (np.random.rand(8, 8) * 255).astype(np.uint8)
        >>> norm_u = normalize(raw_u)

    Example:
        >>> # xdoctest: +REQUIRES(module:kwimage)
        >>> import kwimage
        >>> arr = kwimage.grab_test_image('lowcontrast')
        >>> arr = kwimage.ensure_float01(arr)
        >>> norms = {}
        >>> norms['arr'] = arr.copy()
        >>> norms['linear'] = normalize(arr, mode='linear')
        >>> norms['sigmoid'] = normalize(arr, mode='sigmoid')
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.figure(fnum=1, doclf=True)
        >>> pnum_ = kwplot.PlotNums(nSubplots=len(norms))
        >>> for key, img in norms.items():
        >>>     kwplot.imshow(img, pnum=pnum_(), title=key)

    Benchmark:
        # Our method is faster than standard in-line implementations.

        import timerit
        ti = timerit.Timerit(100, bestof=10, verbose=2, unit='ms')
        arr = kwimage.grab_test_image('lowcontrast', dsize=(512, 512))

        print('--- uint8 ---')
        arr = ensure_float01(arr)
        out = arr.copy()
        for timer in ti.reset('naive1-float'):
            with timer:
                (arr - arr.min()) / (arr.max() - arr.min())

        import timerit
        for timer in ti.reset('simple-float'):
            with timer:
                max_ = arr.max()
                min_ = arr.min()
                result = (arr - min_) / (max_ - min_)

        for timer in ti.reset('normalize-float'):
            with timer:
                normalize(arr)

        for timer in ti.reset('normalize-float-inplace'):
            with timer:
                normalize(arr, out=out)

        print('--- float ---')
        arr = ensure_uint255(arr)
        out = arr.copy()
        for timer in ti.reset('naive1-uint8'):
            with timer:
                (arr - arr.min()) / (arr.max() - arr.min())

        import timerit
        for timer in ti.reset('simple-uint8'):
            with timer:
                max_ = arr.max()
                min_ = arr.min()
                result = (arr - min_) / (max_ - min_)

        for timer in ti.reset('normalize-uint8'):
            with timer:
                normalize(arr)

        for timer in ti.reset('normalize-uint8-inplace'):
            with timer:
                normalize(arr, out=out)

    Ignore:
        globals().update(xdev.get_func_kwargs(normalize))
    """
    if out is None:
        out = arr.copy()

    # TODO:
    # - [ ] Parametarize new_min / new_max values
    #     - [ ] infer from datatype
    #     - [ ] explicitly given
    new_min = 0.0
    if arr.dtype.kind in ('i', 'u'):
        # Need a floating point workspace
        float_out = out.astype(np.float32)
        new_max = float(np.iinfo(arr.dtype).max)
    elif arr.dtype.kind == 'f':
        float_out = out
        new_max = 1.0
    else:
        raise NotImplementedError

    # TODO:
    # - [ ] Parametarize old_min / old_max strategies
    #     - [ ] explicitly given min and max
    #     - [ ] raw-naive min and max inference
    #     - [ ] outlier-aware min and max inference
    old_min = float_out.min()
    old_max = float_out.max()

    old_span = old_max - old_min
    new_span = new_max - new_min

    if mode == 'linear':
        # linear case
        # out = (arr - old_min) * (new_span / old_span) + new_min
        factor = 1.0 if old_span == 0 else (new_span / old_span)
        if old_min != 0:
            float_out -= old_min
    elif mode == 'sigmoid':
        # nonlinear case
        # out = new_span * sigmoid((arr - beta) / alpha) + new_min
        from scipy.special import expit as sigmoid
        if beta is None:
            # should center the desired distribution to visualize on zero
            beta = old_max - old_min

        if alpha is None:
            # division factor
            # from scipy.special import logit
            # alpha = max(abs(old_min - beta), abs(old_max - beta)) / logit(0.998)
            # This chooses alpha such the original min/max value will be pushed
            # towards -1 / +1.
            alpha = max(abs(old_min - beta), abs(old_max - beta)) / 6.212606
        energy = float_out
        energy -= beta
        energy /= alpha
        # Ideally the data of interest is roughly in the range (-6, +6)
        float_out = sigmoid(energy, out=float_out)
        factor = new_span
    else:
        raise KeyError(mode)

    # Stretch / shift to the desired output range
    if factor != 1:
        float_out *= factor

    if new_min != 0:
        float_out += new_min

    if float_out is not out:
        out[:] = float_out.astype(out.dtype)
    return out


# def argsort_threshold(arr, threshold=None, num_top=None, descending=False):
#     """
#     TODO: Cleanup

#     Find all indexes over a threshold, but always return at least the
#     `num_top`.
#     """
#     import kwarray

#     # Find the "best" indices and their scores
#     sortx = arr.argsort(descending=descending)
#     sorted_arr = arr[sortx]
#     # Mark any index "better" than the score threshold
#     if descending:
#         flags = sorted_arr > threshold
#     else:
#         flags = sorted_arr < threshold

#     if num_top is not None:
#         # Always return at least `num_top`
#         flags[0:num_top] = True

#         fallback_thresh = sorted_arr[num_top - 1]
#         threshold = min(fallback_thresh, threshold)

#     top_inds = sortx[flags]
#     return top_inds



import numpy as np
import ubelt as ub


class FlatIndexer(ub.NiceRepr):
    """
    Creates a flat "view" of a jagged nested indexable object.
    Only supports one offset level.

    Args:
        lens (list): a list of the lengths of the nested objects.

    Doctest:
        >>> self = FlatIndexer([1, 2, 3])
        >>> len(self)
        >>> self.unravel(4)
        >>> self.ravel(2, 1)
    """
    def __init__(self, lens):
        self.lens = lens
        self.cums = np.cumsum(lens)

    @classmethod
    def fromlist(cls, items):
        """
        Convenience method to create a :class:`FlatIndexer` from the list of
        items itself instead of the array of lengths.

        Args:
            items (List[list]): a list of the lists you want to flat index over

        Returns:
            FlatIndexer
        """
        lens = list(map(len, items))
        return cls(lens)

    def __len__(self):
        return self.cums[-1] if len(self.cums) else 0

    def unravel(self, index):
        """
        Args:
            index : raveled index

        Returns:
            Tuple[int, int]: outer and inner indices
        """
        outer = np.where(self.cums > index)[0][0]
        base = self.cums[outer] - self.lens[outer]
        inner = index - base
        return (outer, inner)

    def ravel(self, outer, inner):
        """
        Args:
            outer: index into outer list
            inner: index into the list referenced by outer

        Returns:
            index: the raveled index
        """
        base = self.cums[outer] - self.lens[outer]
        return base + inner



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import ubelt as ub
import itertools as it
from collections import OrderedDict


def setcover(candidate_sets_dict, items=None, set_weights=None,
             item_values=None, max_weight=None, algo='approx'):
    """
    Finds a feasible solution to the minimum weight maximum value set cover.
    The quality and runtime of the solution will depend on the backend
    algorithm selected.

    Args:
        candidate_sets_dict (Dict[Hashable, List[Hashable]]):
            a dictionary where keys are the candidate set ids and each value is
            a candidate cover set.

        items (Hashable, optional): the set of all items to be covered,
            if not specified, it is infered from the candidate cover sets

        set_weights (Dict, optional): maps candidate set ids to a cost
            for using this candidate cover in the solution. If not specified
            the weight of each candiate cover defaults to 1.

        item_values (Dict, optional): maps each item to a value we get for
            returning this item in the solution. If not specified the value
            of each item defaults to 1.

        max_weight (float): if specified, the total cost of the returned cover
            is constrained to be less than this number.

        algo (str): specifies which algorithm to use. Can either be
            'approx' for the greedy solution or 'exact' for the globally
            optimal solution. Note the 'exact' algorithm solves an
            integer-linear-program, which can be very slow and requires
            the `pulp` package to be installed.

    Returns:
        Dict: a subdict of candidate_sets_dict containing the chosen solution.

    Example:
        >>> candidate_sets_dict = {
        >>>     'a': [1, 2, 3, 8, 9, 0],
        >>>     'b': [1, 2, 3, 4, 5],
        >>>     'c': [4, 5, 7],
        >>>     'd': [5, 6, 7],
        >>>     'e': [6, 7, 8, 9, 0],
        >>> }
        >>> greedy_soln = setcover(candidate_sets_dict, algo='greedy')
        >>> print('greedy_soln = {}'.format(ub.repr2(greedy_soln, nl=0)))
        greedy_soln = {'a': [1, 2, 3, 8, 9, 0], 'c': [4, 5, 7], 'd': [5, 6, 7]}
        >>> # xdoc: +REQUIRES(module:pulp)
        >>> exact_soln = setcover(candidate_sets_dict, algo='exact')
        >>> print('exact_soln = {}'.format(ub.repr2(exact_soln, nl=0)))
        exact_soln = {'b': [1, 2, 3, 4, 5], 'e': [6, 7, 8, 9, 0]}
    """
    if algo in ['approx', 'greedy']:
        return _setcover_greedy_old(candidate_sets_dict, items=items,
                                    set_weights=set_weights,
                                    item_values=item_values,
                                    max_weight=max_weight)
    elif algo in ['exact', 'ilp']:
        return _setcover_ilp(candidate_sets_dict, items=items,
                             set_weights=set_weights, item_values=item_values,
                             max_weight=max_weight)
    else:
        raise KeyError(algo)


def _setcover_greedy_old(candidate_sets_dict, items=None, set_weights=None,
                         item_values=None, max_weight=None):
    """
    Benchmark:
        items = np.arange(10000)
        candidate_sets_dict = {}
        for i in range(1000):
            candidate_sets_dict[i] = np.random.choice(items, 200).tolist()

        _setcover_greedy_new(candidate_sets_dict) == _setcover_greedy_old(candidate_sets_dict)
        _ = nh.util.profile_onthefly(_setcover_greedy_new)(candidate_sets_dict)
        _ = nh.util.profile_onthefly(_setcover_greedy_old)(candidate_sets_dict)

        import ubelt as ub
        for timer in ub.Timerit(3, bestof=1, label='time'):
            with timer:
                len(_setcover_greedy_new(candidate_sets_dict))

        import ubelt as ub
        for timer in ub.Timerit(3, bestof=1, label='time'):
            with timer:
                len(_setcover_greedy_old(candidate_sets_dict))
    """
    solution_cover = {}

    if len(candidate_sets_dict) == 0:
        # O(1) optimal solution, we did it!
        return solution_cover

    # If candset_weights or item_values not given use the length as defaults
    if set_weights is None:
        get_weight = len
    else:
        def get_weight(solution_cover):
            return sum(set_weights[key] for key in solution_cover.keys())
    if item_values is None:
        get_value = len
    else:
        def get_value(vals):
            return sum(item_values[v] for v in vals)
    if max_weight is None:
        max_weight = get_weight(candidate_sets_dict)

    avail_covers = {key: set(val) for key, val in candidate_sets_dict.items()}
    avail_keys, avail_vals = zip(*sorted(avail_covers.items()))
    avail_keys = list(avail_keys)
    avail_vals = list(avail_vals)

    # While we still need covers
    while get_weight(solution_cover) < max_weight and len(avail_keys) > 0:
        # Find candiate set with the most uncovered items
        uncovered_values = list(map(get_value, avail_vals))
        chosen_idx = ub.argmax(uncovered_values)
        if uncovered_values[chosen_idx] <= 0:
            # needlessly adding value-less items
            break
        chosen_key = avail_keys[chosen_idx]
        # Add values in this key to the cover
        chosen_set = avail_covers[chosen_key]
        solution_cover[chosen_key] = candidate_sets_dict[chosen_key]
        # Remove chosen set from available options and covered items
        # from remaining available sets
        del avail_keys[chosen_idx]
        del avail_vals[chosen_idx]
        for vals in avail_vals:
            vals.difference_update(chosen_set)
    return solution_cover


def _setcover_greedy_new(candidate_sets_dict, items=None, set_weights=None,
                         item_values=None, max_weight=None):
    """
    Implements Johnson's / Chvatal's greedy set-cover approximation algorithms.

    The approximation gaurentees depend on specifications of set weights and
    item values

    Running time:
        N = number of universe items
        C = number of candidate covering sets

        Worst case running time is: O(C^2 * CN)
            (note this is via simple analysis, the big-oh might be better)

    Set Cover: log(len(items) + 1) approximation algorithm
    Weighted Maximum Cover: 1 - 1/e == .632 approximation algorithm
    Generalized maximum coverage is not implemented

    References:
        https://en.wikipedia.org/wiki/Maximum_coverage_problem

    Notes:
        # pip install git+git://github.com/tangentlabs/django-oscar.git#egg=django-oscar.
        # TODO: wrap https://github.com/martin-steinegger/setcover/blob/master/SetCover.cpp
        # pip install SetCoverPy
        # This is actually much slower than my implementation
        from SetCoverPy import setcover
        g = setcover.SetCover(full_overlaps, cost=np.ones(len(full_overlaps)))
        g.greedy()
        keep = np.where(g.s)[0]

    Example:
        >>> candidate_sets_dict = {
        >>>     'a': [1, 2, 3, 8, 9, 0],
        >>>     'b': [1, 2, 3, 4, 5],
        >>>     'c': [4, 5, 7],
        >>>     'd': [5, 6, 7],
        >>>     'e': [6, 7, 8, 9, 0],
        >>> }
        >>> greedy_soln = _setcover_greedy_new(candidate_sets_dict)
        >>> #print(repr(greedy_soln))
        ...
        >>> print('greedy_soln = {}'.format(ub.repr2(greedy_soln, nl=0)))
        greedy_soln = {'a': [1, 2, 3, 8, 9, 0], 'c': [4, 5, 7], 'd': [5, 6, 7]}

    Example:
        >>> candidate_sets_dict = {
        >>>     'a': [1, 2, 3, 8, 9, 0],
        >>>     'b': [1, 2, 3, 4, 5],
        >>>     'c': [4, 5, 7],
        >>>     'd': [5, 6, 7],
        >>>     'e': [6, 7, 8, 9, 0],
        >>> }
        >>> items = list(set(it.chain(*candidate_sets_dict.values())))
        >>> set_weights = {i: 1 for i in candidate_sets_dict.keys()}
        >>> item_values = {e: 1 for e in items}
        >>> greedy_soln = _setcover_greedy_new(candidate_sets_dict,
        >>>                             item_values=item_values,
        >>>                             set_weights=set_weights)
        >>> print('greedy_soln = {}'.format(ub.repr2(greedy_soln, nl=0)))
        greedy_soln = {'a': [1, 2, 3, 8, 9, 0], 'c': [4, 5, 7], 'd': [5, 6, 7]}

    Example:
        >>> candidate_sets_dict = {}
        >>> greedy_soln = _setcover_greedy_new(candidate_sets_dict)
        >>> print('greedy_soln = {}'.format(ub.repr2(greedy_soln, nl=0)))
        greedy_soln = {}
    """
    if len(candidate_sets_dict) == 0:
        # O(1) optimal solution, we did it!
        return {}

    solution_cover = {}
    solution_weight = 0

    if items is None:
        items = list(set(it.chain(*candidate_sets_dict.values())))

    # Inverted index
    item_to_keys = {item: set() for item in items}
    # This is actually a fair bit faster than the non-comprehension version
    [item_to_keys[item].add(key)
     for key, vals in candidate_sets_dict.items()
     for item in vals]

    # If set_weights or item_values not given use the length as defaults
    if set_weights is None:
        get_weight = len
    else:
        # TODO: we can improve this with bookkeeping
        def get_weight(solution_cover):
            return sum(set_weights[key] for key in solution_cover.keys())

    if item_values is None:
        get_value = len
    else:
        def get_value(vals):
            return sum(item_values[v] for v in vals)
    if max_weight is None:
        max_weight = get_weight(candidate_sets_dict)

    avail_covers = OrderedDict([
        (key, set(vals))
        for key, vals in sorted(candidate_sets_dict.items())
    ])
    avail_totals = OrderedDict([
        (key, get_value(vals))
        for key, vals in avail_covers.items()
    ])

    print('avail_covers = {}'.format(ub.repr2(avail_covers, nl=1)))
    print('avail_totals = {}'.format(ub.repr2(avail_totals, nl=1)))

    # While we still need covers
    while solution_weight < max_weight and len(avail_covers) > 0:
        # Find candiate set with the most valuable uncovered items
        chosen_key = ub.argmax(avail_totals)
        if avail_totals[chosen_key] <= 0:
            # needlessly adding value-less covering set
            break

        print('-----')
        print('CHOOSE COVER SET = {!r}'.format(chosen_key))

        # Add values in this key to the cover
        chosen_items = avail_covers[chosen_key]
        solution_cover[chosen_key] = candidate_sets_dict[chosen_key]

        # Update the solution weight
        chosen_weight = (1 if set_weights is None else set_weights[chosen_key])
        solution_weight += chosen_weight

        # Remove chosen covering set from available options
        del avail_covers[chosen_key]
        del avail_totals[chosen_key]

        # For each chosen item, find the other sets that it belongs to
        modified_keys = set()
        for item in chosen_items:
            # Update the inverted index
            new_keys = item_to_keys[item]
            new_keys.remove(chosen_key)
            item_to_keys[item] = new_keys
            # And mark the non-chosen reamining cover sets as modified
            modified_keys.update(new_keys)
        # Then update and recompute the value of the modified sets
        for key in modified_keys:
            avail_covers[key].difference_update(chosen_items)
            newval = get_value(avail_covers[key])
            avail_totals[key] = newval

        print('avail_covers = {}'.format(ub.repr2(avail_covers, nl=1)))
        print('avail_totals = {}'.format(ub.repr2(avail_totals, nl=1)))

    print('solution_cover = {!r}'.format(solution_cover))
    return solution_cover


def _setcover_ilp(candidate_sets_dict, items=None, set_weights=None,
                  item_values=None, max_weight=None, verbose=False):
    """
    Set cover / Weighted Maximum Cover exact algorithm using an integer linear
    program.

    TODO:
        - [ ] Use CPLEX solver if available

    https://en.wikipedia.org/wiki/Maximum_coverage_problem

    Example:
        >>> # xdoc: +REQUIRES(module:pulp)
        >>> candidate_sets_dict = {}
        >>> exact_soln = _setcover_ilp(candidate_sets_dict)
        >>> print('exact_soln = {}'.format(ub.repr2(exact_soln, nl=0)))
        exact_soln = {}

    Example:
        >>> # xdoc: +REQUIRES(module:pulp)
        >>> candidate_sets_dict = {
        >>>     'a': [1, 2, 3, 8, 9, 0],
        >>>     'b': [1, 2, 3, 4, 5],
        >>>     'c': [4, 5, 7],
        >>>     'd': [5, 6, 7],
        >>>     'e': [6, 7, 8, 9, 0],
        >>> }
        >>> items = list(set(it.chain(*candidate_sets_dict.values())))
        >>> set_weights = {i: 1 for i in candidate_sets_dict.keys()}
        >>> item_values = {e: 1 for e in items}
        >>> exact_soln1 = _setcover_ilp(candidate_sets_dict,
        >>>                             item_values=item_values,
        >>>                             set_weights=set_weights)
        >>> exact_soln2 = _setcover_ilp(candidate_sets_dict)
        >>> assert exact_soln1 == exact_soln2
    """
    try:
        import pulp
    except ImportError:
        print('ERROR: must install pulp to use ILP setcover solver')
        raise

    if len(candidate_sets_dict) == 0:
        return {}

    if items is None:
        items = list(set(it.chain(*candidate_sets_dict.values())))

    if item_values is None and set_weights is None and max_weight is None:
        # This is the most basic set cover problem
        # Formulate integer program
        prob = pulp.LpProblem("Set Cover", pulp.LpMinimize)
        # Solution variable indicates if set it chosen or not
        set_indices = candidate_sets_dict.keys()
        x = pulp.LpVariable.dicts(name='x', indexs=set_indices,
                                  lowBound=0, upBound=1, cat=pulp.LpInteger)
        # minimize the number of sets
        prob.objective = sum(x[i] for i in set_indices)
        # subject to
        for e in items:
            # each element is covered
            containing_sets = [i for i in set_indices if e in candidate_sets_dict[i]]
            prob.add(sum(x[i] for i in containing_sets) >= 1)
        # Solve using with solver like CPLEX, GLPK, or SCIP.
        #pulp.CPLEX().solve(prob)
        pulp.PULP_CBC_CMD().solve(prob)
        # Read solution
        solution_keys = [i for i in set_indices if x[i].varValue]
        solution_cover = {i: candidate_sets_dict[i] for i in solution_keys}
        # Print summary
        if verbose:
            print(prob)
            print('OPT:')
            print('\n'.join(['    %s = %s' % (x[i].name, x[i].varValue) for i in set_indices]))
            print('solution_cover = %r' % (solution_cover,))
    else:
        if set_weights is None:
            set_weights = {i: 1 for i in candidate_sets_dict.keys()}
        if item_values is None:
            item_values = {e: 1 for e in items}
        if max_weight is None:
            max_weight = sum(set_weights[i] for i in candidate_sets_dict.keys())
        prob = pulp.LpProblem("Maximum Cover", pulp.LpMaximize)
        # Solution variable indicates if set it chosen or not
        item_indicies = items
        set_indices = candidate_sets_dict.keys()
        x = pulp.LpVariable.dicts(name='x', indexs=set_indices,
                                  lowBound=0, upBound=1, cat=pulp.LpInteger)
        y = pulp.LpVariable.dicts(name='y', indexs=item_indicies,
                                  lowBound=0, upBound=1, cat=pulp.LpInteger)
        r = pulp.LpVariable.dicts(name='r', indexs=item_indicies)
        # maximize the value of the covered items
        primary_objective = sum(item_values[e] * y[e] for e in item_indicies)
        # minimize the number of sets used (make sure it does not influence the chosen primary objective)
        # This is only possible when values are non-negative
        # TODO: minimize redundency
        min_influence = min(item_values.values())
        secondary_weight = min_influence / (1.1 * len(set_indices))
        secondary_objective = (sum(-x[i] for i in set_indices)) * secondary_weight
        #
        prob.objective = primary_objective + secondary_objective
        # subject to
        # no more than the maximum weight
        prob.add(sum(x[i] * set_weights[i] for i in set_indices) <= max_weight)
        # If an item is chosen than at least one set containing it is chosen
        for e in item_indicies:
            containing_sets = [i for i in set_indices if e in candidate_sets_dict[i]]
            if len(containing_sets) > 0:
                prob.add(sum(x[i] for i in containing_sets) >= y[e])
                # record number of times each item is covered
                prob.add(sum(x[i] for i in containing_sets) == r[e])
        # Solve using with solver like CPLEX, GLPK, or SCIP.
        #pulp.CPLEX().solve(prob)
        pulp.PULP_CBC_CMD().solve(prob)
        # Read solution
        solution_keys = [i for i in set_indices if x[i].varValue]
        solution_cover = {i: candidate_sets_dict[i] for i in solution_keys}
        # Print summary
        if verbose:
            print(prob)
            print('OPT:')
            print('\n'.join(['    %s = %s' % (x[i].name, x[i].varValue) for i in set_indices]))
            print('\n'.join(['    %s = %s' % (y[i].name, y[i].varValue) for i in item_indicies]))
            print('solution_cover = %r' % (solution_cover,))
    return solution_cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
from os.path import dirname, exists, join
from setuptools import setup
from setuptools import find_packages


def parse_version(fpath):
    """
    Statically parse the version number from a python file
    """
    import ast
    if not exists(fpath):
        raise ValueError('fpath={!r} does not exist'.format(fpath))
    with open(fpath, 'r') as file_:
        sourcecode = file_.read()
    pt = ast.parse(sourcecode)
    class VersionVisitor(ast.NodeVisitor):
        def visit_Assign(self, node):
            for target in node.targets:
                if getattr(target, 'id', None) == '__version__':
                    self.version = node.value.s
    visitor = VersionVisitor()
    visitor.visit(pt)
    return visitor.version


def parse_description():
    """
    Parse the description in the README file

    CommandLine:
        pandoc --from=markdown --to=rst --output=README.rst README.md
        python -c "import setup; print(setup.parse_description())"
    """
    readme_fpath = join(dirname(__file__), 'README.rst')
    # This breaks on pip install, so check that it exists.
    if exists(readme_fpath):
        with open(readme_fpath, 'r') as f:
            text = f.read()
        return text
    return ''


def parse_requirements(fname='requirements.txt'):
    """
    Parse the package dependencies listed in a requirements file but strips
    specific versioning information.

    TODO:
        perhaps use https://github.com/davidfischer/requirements-parser instead

    CommandLine:
        python -c "import setup; print(setup.parse_requirements())"
    """
    from os.path import exists
    import re
    require_fpath = fname

    def parse_line(line):
        """
        Parse information from a line in a requirements text file
        """
        if line.startswith('-r '):
            # Allow specifying requirements in other files
            target = line.split(' ')[1]
            for info in parse_require_file(target):
                yield info
        elif line.startswith('-e '):
            info = {}
            info['package'] = line.split('#egg=')[1]
            yield info
        else:
            # Remove versioning from the package
            pat = '(' + '|'.join(['>=', '==', '>']) + ')'
            parts = re.split(pat, line, maxsplit=1)
            parts = [p.strip() for p in parts]

            info = {}
            info['package'] = parts[0]
            if len(parts) > 1:
                op, rest = parts[1:]
                if ';' in rest:
                    # Handle platform specific dependencies
                    # http://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-platform-specific-dependencies
                    version, platform_deps = map(str.strip, rest.split(';'))
                    info['platform_deps'] = platform_deps
                else:
                    version = rest  # NOQA
                info['version'] = (op, version)
            yield info

    def parse_require_file(fpath):
        with open(fpath, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line and not line.startswith('#'):
                    for info in parse_line(line):
                        yield info

    # This breaks on pip install, so check that it exists.
    packages = []
    if exists(require_fpath):
        for info in parse_require_file(require_fpath):
            package = info['package']
            if not sys.version.startswith('3.4'):
                # apparently package_deps are broken in 3.4
                platform_deps = info.get('platform_deps')
                if platform_deps is not None:
                    package += ';' + platform_deps
            packages.append(package)
    return packages


def native_mb_python_tag(plat_impl=None, version_info=None):
    """
    Example:
        >>> print(native_mb_python_tag())
        >>> print(native_mb_python_tag('PyPy', (2, 7)))
        >>> print(native_mb_python_tag('CPython', (3, 8)))
    """
    if plat_impl is None:
        import platform
        plat_impl = platform.python_implementation()

    if version_info is None:
        import sys
        version_info = sys.version_info

    major, minor = version_info[0:2]
    ver = '{}{}'.format(major, minor)

    if plat_impl == 'CPython':
        # TODO: get if cp27m or cp27mu
        impl = 'cp'
        if ver == '27':
            IS_27_BUILT_WITH_UNICODE = True  # how to determine this?
            if IS_27_BUILT_WITH_UNICODE:
                abi = 'mu'
            else:
                abi = 'm'
        else:
            if sys.version_info[:2] >= (3, 8):
                # bpo-36707: 3.8 dropped the m flag
                abi = ''
            else:
                abi = 'm'
        mb_tag = '{impl}{ver}-{impl}{ver}{abi}'.format(**locals())
    elif plat_impl == 'PyPy':
        abi = ''
        impl = 'pypy'
        ver = '{}{}'.format(major, minor)
        mb_tag = '{impl}-{ver}'.format(**locals())
    else:
        raise NotImplementedError(plat_impl)
    return mb_tag


NAME = 'kwarray'
VERSION = version = parse_version('kwarray/__init__.py')  # needs to be a global var for git tags

if __name__ == '__main__':
    setup(
        name=NAME,
        version=VERSION,
        author='Kitware Inc., Jon Crall',
        author_email='kitware@kitware.com, jon.crall@kitware.com',
        url='https://gitlab.kitware.com/computer-vision/kwarray',
        long_description=parse_description(),
        long_description_content_type='text/x-rst',
        install_requires=parse_requirements('requirements/runtime.txt'),
        extras_require={
            'all': parse_requirements('requirements.txt'),
            'tests': parse_requirements('requirements/tests.txt'),
            'optional': parse_requirements('requirements/optional.txt'),
        },
        license='Apache 2',
        packages=find_packages(include='kwarray.*'),
        classifiers=[
            # List of classifiers available at:
            # https://pypi.python.org/pypi?%3Aaction=list_classifiers
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering',
            'Topic :: Software Development',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Utilities',
            # This should be interpreted as Apache License v2.0
            'License :: OSI Approved :: Apache Software License',
            # Supported Python versions
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
        ],
    )



# -*- coding: utf-8 -*-
"""
The ArrayAPI is a common API that works exactly the same on both torch.Tensors
and numpy.ndarrays.


The ArrayAPI is a combination of efficiency and convinience. It is convinient
because you can just use an operation directly, it will type check the data,
and apply the appropriate method. But it is also efficient because it can be
used with minimal type checking by accessing a type-specific backend.

For example, you can do:

.. code:: python

    impl = kwarray.ArrayAPI.coerce(data)

And then impl will give you direct access to the appropriate methods without
any type checking overhead.  e..g. ``impl.<op-you-want>(data)``

But you can also do ``kwarray.ArrayAPI.<op-you-want>(data)`` on anything and it
will do type checking and then do the operation you want.

Example:
    >>> # xdoctest: +REQUIRES(module:torch)
    >>> import torch
    >>> import numpy as np
    >>> data1 = torch.rand(10, 10)
    >>> data2 = data1.numpy()
    >>> # Method 1: grab the appropriate sub-impl
    >>> impl1 = ArrayAPI.impl(data1)
    >>> impl2 = ArrayAPI.impl(data2)
    >>> result1 = impl1.sum(data1, axis=0)
    >>> result2 = impl2.sum(data2, axis=0)
    >>> res1_np = ArrayAPI.numpy(result1)
    >>> res2_np = ArrayAPI.numpy(result2)
    >>> print('res1_np = {!r}'.format(res1_np))
    >>> print('res2_np = {!r}'.format(res2_np))
    >>> assert np.allclose(res1_np, res2_np)
    >>> # Method 2: choose the impl on the fly
    >>> result1 = ArrayAPI.sum(data1, axis=0)
    >>> result2 = ArrayAPI.sum(data2, axis=0)
    >>> res1_np = ArrayAPI.numpy(result1)
    >>> res2_np = ArrayAPI.numpy(result2)
    >>> print('res1_np = {!r}'.format(res1_np))
    >>> print('res2_np = {!r}'.format(res2_np))
    >>> assert np.allclose(res1_np, res2_np)

Example:
    >>> # xdoctest: +REQUIRES(module:torch)
    >>> import torch
    >>> import numpy as np
    >>> data1 = torch.rand(10, 10)
    >>> data2 = data1.numpy()
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import ubelt as ub
import six
from functools import partial
from distutils.version import LooseVersion


try:
    import torch
except Exception:
    torch = None
    _TORCH_HAS_BOOL_COMP = False
else:
    _TORCH_HAS_BOOL_COMP = LooseVersion(torch.__version__) >= LooseVersion('1.2.0')


def _get_funcname(func):
    if six.PY2:
        try:
            return func.func_name
        except AttributeError:
            return func.__name__
    else:
        return func.__name__


class _ImplRegistry(object):
    def __init__(self):
        self.registered = {
            'torch': {},
            'numpy': {},
            'api': {},
        }

    def _register(self, func, func_type, impl):
        func_name = _get_funcname(func)

        assert func_type in {
            'data_func',           # methods where the first argument is an array
            'array_sequence',  # methods that take a sequence of arrays
            'shape_creation',  # methods that allocate new data via shape
        }
        self.registered[impl][func_name] = {
            'func': func,
            'func_name': func_name,
            'func_type': func_type,
        }
        return staticmethod(func)

    def _implmethod(self, func=None, func_type='data_func', impl=None):
        _register = partial(self._register, func_type=func_type, impl=impl)
        if func is None:
            return _register
        else:
            return _register(func)

    def _apimethod(self, key=None, func_type='data_func'):
        """
        Creates wrapper for a "data method" --- i.e. a ArrayAPI function that has
        only one main argument, which is an array.
        """
        if isinstance(key, six.string_types):
            # numpy_func = self.registered['numpy'][key]['func']
            # torch_func = self.registered['torch'][key]['func']
            numpy_func = getattr(NumpyImpls, key)
            if torch is not None:
                torch_func = getattr(TorchImpls, key)
            def func(data, *args, **kwargs):
                if torch is not None and torch.is_tensor(data):
                    return torch_func(data, *args, **kwargs)
                elif isinstance(data, np.ndarray):
                    return numpy_func(data, *args, **kwargs)
                elif isinstance(data, (list, tuple)):
                    return numpy_func(np.asarray(data), *args, **kwargs)
                else:
                    raise TypeError('unknown type {}'.format(type(data)))
            func.__name__ = str(key)  # the str wrap is for python2
        else:
            func = key
        _register = partial(self._register, func_type=func_type, impl='api')
        if func is None:
            return _register
        else:
            return _register(func)

    def _ensure_datamethods_names_are_registered(self):
        """
        Checks to make sure all methods are implemented in both
        torch and numpy implementations as well as exposed in the ArrayAPI.
        """
        # Check that we didn't forget to add anything
        # api_names = {
        #     key for key, value in ArrayAPI.__dict__.items()
        #     if isinstance(value, staticmethod)
        # }
        # import xdev
        # xdev.fix_embed_globals()
        numpy_names = set(self.registered['numpy'].keys())
        torch_names = set(self.registered['torch'].keys())
        api_names = set(self.registered['api'].keys())

        universe = (torch_names | numpy_names)

        grouped = ub.group_items(self.registered['numpy'].values(), lambda x: x['func_type'])
        shape_creation_methods = {item['func_name'] for item in grouped['shape_creation']}

        missing_torch = universe - torch_names
        missing_numpy = universe - numpy_names
        missing_api = universe - api_names - shape_creation_methods

        if missing_numpy or missing_torch:
            message = ub.codeblock(
                '''
                missing_torch = {}
                missing_numpy = {}
                ''').format(missing_torch, missing_numpy)
            raise AssertionError(message)

        if missing_api:
            message = ub.codeblock(
                '''
                WARNING (these will be implicitly created):
                missing_api = {}
                '''.format(missing_api))
            print(message)
            import warnings
            warnings.warn(message)

        # Insert any missing functions into the API implicitly
        autogen = []
        for func_name in missing_api:
            func_type = self.registered['numpy'][func_name]['func_type']
            autogen.append("{} = _apimethod('{}', func_type='{}')".format(func_name, func_name, func_type))
            implicitmethod = self._apimethod(func_name, func_type=func_type)
            setattr(ArrayAPI, func_name, implicitmethod)

        for func_name in universe:
            infos = []
            try:
                for impl in ['numpy', 'torch', 'api']:
                    info = self.registered[impl][func_name]
                    info.pop('func')
                    infos.append(info)
                assert ub.allsame(infos), 'fix {}'.format(infos)
            except KeyError:
                pass

        if autogen:
            print('Please explicitly add the following code:')
            print('\n'.join(autogen))

# alias staticmethod to make it easier to keep track of which IMPL we are in

_REGISTERY = _ImplRegistry()
_torchmethod = partial(_REGISTERY._implmethod, impl='torch')
_numpymethod = partial(_REGISTERY._implmethod, impl='numpy')
_apimethod = _REGISTERY._apimethod


class TorchImpls(object):
    """
    Torch backend for the ArrayAPI API
    """

    is_tensor = True
    is_numpy = False

    @_torchmethod(func_type='array_sequence')
    def cat(datas, axis=-1):
        return torch.cat(datas, dim=axis)

    @_torchmethod(func_type='array_sequence')
    def hstack(datas):
        """
        Concatenates along axis=0 if inputs are are 1-D otherwise axis=1

        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> datas1 = [torch.arange(10), torch.arange(10)]
            >>> datas2 = [d.numpy() for d in datas1]
            >>> ans1 = TorchImpls.hstack(datas1)
            >>> ans2 = NumpyImpls.hstack(datas2)
            >>> assert np.all(ans1.numpy() == ans2)
        """
        axis = 1
        if len(datas[0].shape) == 1:
            axis = 0
        return torch.cat(datas, dim=axis)

    @_torchmethod(func_type='array_sequence')
    def vstack(datas):
        """
        Ensures that inputs datas are at least 2D (prepending a dimension of 1)
        and then concats along axis=0.

        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> datas1 = [torch.arange(10), torch.arange(10)]
            >>> datas2 = [d.numpy() for d in datas1]
            >>> ans1 = TorchImpls.vstack(datas1)
            >>> ans2 = NumpyImpls.vstack(datas2)
            >>> assert np.all(ans1.numpy() == ans2)
        """
        datas = [TorchImpls.atleast_nd(d, n=2, front=True) for d in datas]
        return torch.cat(datas, dim=0)

    @_torchmethod(func_type='data_func')
    def atleast_nd(arr, n, front=False):
        ndims = len(arr.shape)
        if n is not None and ndims <  n:
            # append the required number of dimensions to the front or back
            if front:
                expander = (None,) * (n - ndims) + (Ellipsis,)
            else:
                expander = (Ellipsis,) + (None,) * (n - ndims)
            arr = arr[expander]
        return arr

    @_torchmethod(func_type='data_func')
    def view(data, *shape):
        data_ = data.view(*shape)
        return data_

    @_torchmethod(func_type='data_func')
    def take(data, indices, axis=None):
        if not torch.is_tensor(indices):
            indices = torch.LongTensor(indices).to(data.device)
        if axis is None:
            return data.take(indices)
        else:
            return torch.index_select(data, dim=axis, index=indices)

    @_torchmethod(func_type='data_func')
    def compress(data, flags, axis=None):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> import kwarray
            >>> data = torch.rand(10, 4, 2)
            >>> impl = kwarray.ArrayAPI.coerce(data)

            >>> axis = 0
            >>> flags = (torch.arange(data.shape[axis]) % 2) == 0
            >>> out = impl.compress(data, flags, axis=axis)
            >>> assert tuple(out.shape) == (5, 4, 2)

            >>> axis = 1
            >>> flags = (torch.arange(data.shape[axis]) % 2) == 0
            >>> out = impl.compress(data, flags, axis=axis)
            >>> assert tuple(out.shape) == (10, 2, 2)

            >>> axis = 2
            >>> flags = (torch.arange(data.shape[axis]) % 2) == 0
            >>> out = impl.compress(data, flags, axis=axis)
            >>> assert tuple(out.shape) == (10, 4, 1)

            >>> axis = None
            >>> data = torch.rand(10)
            >>> flags = (torch.arange(data.shape[0]) % 2) == 0
            >>> out = impl.compress(data, flags, axis=axis)
            >>> assert tuple(out.shape) == (5,)
        """
        if not torch.is_tensor(flags):
            flags = np.asarray(flags).astype(np.uint8)
            if _TORCH_HAS_BOOL_COMP:
                flags = torch.BoolTensor(flags).to(data.device)
            else:
                flags = torch.ByteTensor(flags).to(data.device)
        if flags.ndimension() != 1:
            raise ValueError('condition must be a 1-d tensor')
        if axis is None:
            return torch.masked_select(data.view(-1), flags)
        else:
            out_shape = list(data.shape)
            out_shape[axis] = int(flags.sum())

            fancy_shape = [1] * len(out_shape)
            fancy_shape[axis] = flags.numel()
            explicit_flags = flags.view(*fancy_shape)

            flat_out = torch.masked_select(data, explicit_flags)
            out = flat_out.view(*out_shape)
            return out

    @_torchmethod(func_type='data_func')
    def tile(data, reps):
        """
        Implement np.tile in torch

        Example:
            >>> # xdoctest: +SKIP
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> data = torch.arange(10)[:, None]
            >>> ans1 = ArrayAPI.tile(data, [1, 2])
            >>> ans2 = ArrayAPI.tile(data.numpy(), [1, 2])
            >>> assert np.all(ans1.numpy() == ans2)

        Doctest:
            >>> # xdoctest: +SKIP
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> shapes = [(3,), (3, 4,), (3, 5, 7), (1,), (3, 1, 3)]
            >>> for shape in shapes:
            >>>     data = torch.rand(*shape)
            >>>     for axis in range(len(shape)):
            >>>         for reps in it.product(*[range(0, 4)] * len(shape)):
            >>>             ans1 = ArrayAPI.tile(data, reps)
            >>>             ans2 = ArrayAPI.tile(data.numpy(), reps)
            >>>             #print('ans1.shape = {!r}'.format(tuple(ans1.shape)))
            >>>             #print('ans2.shape = {!r}'.format(tuple(ans2.shape)))
            >>>             assert np.all(ans1.numpy() == ans2)
        """
        n_prepend = len(reps) - len(data.shape)
        if n_prepend > 0:
            reps = [1] * n_prepend + list(reps)
        if any(r == 0 for r in reps):
            newshape = np.array([r * s for r, s in zip(reps, data.shape)])
            return torch.empty(tuple(newshape), dtype=data.dtype).to(data.device)
        else:
            return data.repeat(reps)

    @_torchmethod(func_type='data_func')
    def repeat(data, repeats, axis=None):
        """
        I'm not actually sure how to implement this efficiently

        Example:
            >>> # xdoctest: +SKIP
            >>> data = torch.arange(10)[:, None]
            >>> ans1 = ArrayAPI.repeat(data, 2, axis=1)
            >>> ans2 = ArrayAPI.repeat(data.numpy(), 2, axis=1)
            >>> assert np.all(ans1.numpy() == ans2)

        Doctest:
            >>> # xdoctest: +SKIP
            >>> shapes = [(3,), (3, 4,), (3, 5, 7)]
            >>> for shape in shapes:
            >>>     data = torch.rand(*shape)
            >>>     for axis in range(len(shape)):
            >>>         for repeats in range(0, 4):
            >>>             ans1 = ArrayAPI.repeat(data, repeats, axis=axis)
            >>>             ans2 = ArrayAPI.repeat(data.numpy(), repeats, axis=axis)
            >>>             assert np.all(ans1.numpy() == ans2)


            ArrayAPI.repeat(data, 2, axis=0)
            ArrayAPI.repeat(data.numpy(), 2, axis=0)

            x = np.array([[1,2],[3,4]])
            np.repeat(x, [1, 2], axis=0)

            ArrayAPI.repeat(data.numpy(), [1, 2])
        """
        raise NotImplementedError
        if False:
            # hack! may not always work
            return np.repeat(data, repeats, axis=axis)
        else:
            if isinstance(repeats, int):
                if axis is None:
                    return data.view(-1).repeat(repeats)
                else:
                    expander = [1] * len(data.shape)
                    expander[axis] = repeats
                    if repeats == 0:
                        newshape = list(data.shape)
                        newshape[axis] = 0
                        return torch.empty(tuple(newshape), dtype=data.dtype).to(data.device)
                    else:
                        return data.repeat(expander)
            else:
                raise NotImplementedError
        pass

    @_torchmethod(func_type='data_func')
    def T(data):
        ndims = data.ndimension()
        if ndims == 2:
            # torch.t can only handle 2 dims
            return data.t()
        else:
            # use permute for compatability
            axes = list(reversed(range(ndims)))
            return data.permute(*axes)

    @_torchmethod(func_type='data_func')
    def transpose(data, axes):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> data1 = torch.rand(2, 3, 5)
            >>> data2 = data1.numpy()
            >>> res1 = ArrayAPI.transpose(data1, (2, 0, 1))
            >>> res2 = ArrayAPI.transpose(data2, (2, 0, 1))
            >>> assert np.all(res1.numpy() == res2)
        """
        return data.permute(axes)

    @_torchmethod(func_type='data_func')
    def numel(data):
        return data.numel()

    # --- Allocators ----

    @_torchmethod(func_type='data_func')
    def full_like(data, fill_value, dtype=None):
        dtype = _torch_dtype_lut().get(dtype, dtype)
        return torch.full_like(data, fill_value, dtype=dtype)

    @_torchmethod(func_type='data_func')
    def empty_like(data, dtype=None):
        dtype = _torch_dtype_lut().get(dtype, dtype)
        return torch.empty_like(data, dtype=dtype)

    @_torchmethod(func_type='data_func')
    def zeros_like(data, dtype=None):
        dtype = _torch_dtype_lut().get(dtype, dtype)
        return torch.zeros_like(data, dtype=dtype)

    @_torchmethod(func_type='data_func')
    def ones_like(data, dtype=None):
        dtype = _torch_dtype_lut().get(dtype, dtype)
        return torch.ones_like(data, dtype=dtype)

    @_torchmethod(func_type='shape_creation')
    def full(shape, fill_value, dtype=float):
        dtype = _torch_dtype_lut().get(dtype, dtype)
        return torch.full(shape, fill_value, dtype=dtype)

    @_torchmethod(func_type='shape_creation')
    def empty(shape, dtype=float):
        dtype = _torch_dtype_lut().get(dtype, dtype)
        return torch.full(shape, dtype=dtype)

    @_torchmethod(func_type='shape_creation')
    def zeros(shape, dtype=float):
        dtype = _torch_dtype_lut().get(dtype, dtype)
        return torch.zeros(shape, dtype=dtype)

    @_torchmethod(func_type='shape_creation')
    def ones(shape, dtype=float):
        dtype = _torch_dtype_lut().get(dtype, dtype)
        return torch.ones(shape, dtype=dtype)

    # -------

    @_torchmethod(func_type='data_func')
    def argmax(data, axis=None):
        return torch.argmax(data, dim=axis)

    @_torchmethod(func_type='data_func')
    def argsort(data, axis=-1, descending=False):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> from kwarray.arrayapi import *  # NOQA
            >>> rng = np.random.RandomState(0)
            >>> data2 = rng.rand(5, 5)
            >>> data1 = torch.from_numpy(data2)
            >>> res1 = ArrayAPI.argsort(data1)
            >>> res2 = ArrayAPI.argsort(data2)
            >>> assert np.all(res1.numpy() == res2)
            >>> res1 = ArrayAPI.argsort(data1, axis=1)
            >>> res2 = ArrayAPI.argsort(data2, axis=1)
            >>> assert np.all(res1.numpy() == res2)
            >>> res1 = ArrayAPI.argsort(data1, axis=1, descending=True)
            >>> res2 = ArrayAPI.argsort(data2, axis=1, descending=True)
            >>> assert np.all(res1.numpy() == res2)
            >>> data2 = rng.rand(5)
            >>> data1 = torch.from_numpy(data2)
            >>> res1 = ArrayAPI.argsort(data1, axis=0, descending=True)
            >>> res2 = ArrayAPI.argsort(data2, axis=0, descending=True)
            >>> assert np.all(res1.numpy() == res2)
        """
        return torch.argsort(data, dim=axis, descending=descending)

    @_torchmethod(func_type='data_func')
    def max(data, axis=None):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> data1 = torch.rand(5, 5, 5, 5, 5, 5)
            >>> data2 = data1.numpy()
            >>> res1 = ArrayAPI.max(data1)
            >>> res2 = ArrayAPI.max(data2)
            >>> assert np.all(res1.numpy() == res2)
            >>> res1 = ArrayAPI.max(data1, axis=(4, 0, 1))
            >>> res2 = ArrayAPI.max(data2, axis=(4, 0, 1))
            >>> assert np.all(res1.numpy() == res2)
            >>> res1 = ArrayAPI.max(data1, axis=(5, -2))
            >>> res2 = ArrayAPI.max(data2, axis=(5, -2))
            >>> assert np.all(res1.numpy() == res2)
        """
        if axis is None:
            return torch.max(data)
        elif isinstance(axis, tuple):
            axis_ = [d if d >= 0 else len(data.shape) + d for d in axis]
            temp = data
            for d in sorted(axis_)[::-1]:
                temp = temp.max(dim=d)[0]
            return temp
        else:
            return torch.max(data, dim=axis)[0]

    @_torchmethod(func_type='data_func')
    def max_argmax(data, axis=None):
        """
        Note: this isn't always gaurenteed to be compatibile with numpy
        if there are equal elements in data. See:
        >>> np.ones(10).argmax()   # xdoctest: +IGNORE_WANT
        0
        >>> torch.ones(10).argmax()   # xdoctest: +IGNORE_WANT
        tensor(9)
        """
        return torch.max(data, dim=axis)

    @_torchmethod
    def maximum(data1, data2, out=None):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> data1 = torch.rand(5, 5)
            >>> data2 = torch.rand(5, 5)
            >>> result1 = TorchImpls.maximum(data1, data2)
            >>> result2 = NumpyImpls.maximum(data1.numpy(), data2.numpy())
            >>> assert np.allclose(result1.numpy(), result2)
        """
        return torch.max(data1, data2, out=out)

    @_torchmethod
    def minimum(data1, data2, out=None):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> data1 = torch.rand(5, 5)
            >>> data2 = torch.rand(5, 5)
            >>> result1 = TorchImpls.minimum(data1, data2)
            >>> result2 = NumpyImpls.minimum(data1.numpy(), data2.numpy())
            >>> assert np.allclose(result1.numpy(), result2)
        """
        return torch.min(data1, data2, out=out)

    # @_torchmethod(func_type='data_func')
    # def matmul(data1, data2, out=None):
    #     return torch.matmul(data1, data2, out=out)
    if torch is not None:
        matmul = _torchmethod(torch.matmul)

    @_torchmethod(func_type='data_func')
    def sum(data, axis=None):
        if axis is None:
            return data.sum()
        else:
            return data.sum(dim=axis)

    @_torchmethod(func_type='data_func')
    def nan_to_num(x, copy=True):
        if copy:
            x = x.clone()
        x[torch.isnan(x)] = 0
        return x

    @_torchmethod(func_type='data_func')
    def copy(data):
        return torch.clone(data)

    if torch is not None:
        log = _torchmethod(torch.log)
        log2 = _torchmethod(torch.log2)
        any = _torchmethod(torch.any)
        all = _torchmethod(torch.all)

    @_torchmethod(func_type='data_func')
    def nonzero(data):
        # torch returns an NxD tensor, whereas numpy returns
        # a D-tuple of N-dimensional arrays.
        return tuple(torch.nonzero(data).t())

    @_torchmethod(func_type='data_func')
    def astype(data, dtype, copy=True):
        dtype = _torch_dtype_lut().get(dtype, dtype)
        data = data.to(dtype)
        if copy:
            data = data.clone()
        return data

    @_torchmethod(func_type='data_func')
    def tensor(data, device=ub.NoParam):
        if device is not ub.NoParam:
            data = data.to(device)
        return data

    @_torchmethod(func_type='data_func')
    def numpy(data):
        return data.data.cpu().numpy()

    @_torchmethod(func_type='data_func')
    def tolist(data):
        return data.data.cpu().numpy().tolist()

    @_torchmethod(func_type='data_func')
    def contiguous(data):
        return data.contiguous()

    @_torchmethod(func_type='data_func')
    def pad(data, pad_width, mode='constant'):
        pad = list(ub.flatten(pad_width[::-1]))
        return torch.nn.functional.pad(data, pad, mode=mode)

    @_torchmethod(func_type='data_func')
    def asarray(data, dtype=None):
        """
        Cast data into a tensor representation

        Example:
            >>> data = np.empty((2, 0, 196, 196), dtype=np.float32)
        """
        if not isinstance(data, torch.Tensor):
            data
            try:
                data =  torch.Tensor(data)
            except RuntimeError:
                if data.size == 0:
                    want_shape = data.shape
                    data_ = torch.empty([0])
                    data = data_.view(want_shape)
        if dtype is not None:
            dtype = _torch_dtype_lut().get(dtype, dtype)
            data = data.to(dtype)
        return data

    ensure = asarray

    @_torchmethod(func_type='data_func')
    def dtype_kind(data):
        """ returns the numpy code for the data type kind """
        if data.dtype.is_floating_point:
            return 'f'
        elif data.dtype == torch.uint8:
            return 'u'
        else:
            return 'i'

    @_torchmethod(func_type='data_func')
    def floor(data, out=None):
        return torch.floor(data, out=out)

    @_torchmethod(func_type='data_func')
    def ceil(data, out=None):
        return torch.ceil(data, out=out)

    @_torchmethod(func_type='data_func')
    def ifloor(data, out=None):
        return torch.floor(data, out=out).int()

    @_torchmethod(func_type='data_func')
    def iceil(data, out=None):
        return torch.ceil(data, out=out).int()

    @_torchmethod(func_type='data_func')
    def round(data, decimals=0, out=None):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> import kwarray
            >>> rng = kwarray.ensure_rng(0)
            >>> np_data = rng.rand(10) * 100
            >>> pt_data = torch.from_numpy(np_data)
            >>> a = kwarray.ArrayAPI.round(np_data)
            >>> b = kwarray.ArrayAPI.round(pt_data)
            >>> assert np.all(a == b.numpy())
            >>> a = kwarray.ArrayAPI.round(np_data, 2)
            >>> b = kwarray.ArrayAPI.round(pt_data, 2)
            >>> assert np.all(a == b.numpy())
        """
        if decimals == 0:
            return torch.round(data, out=out)
        else:
            factor = 10 ** decimals
            result = torch.round(data * factor, out=out)
            result /= factor
            return result

    @_torchmethod(func_type='data_func')
    def iround(data, out=None, dtype=int):
        dtype = _torch_dtype_lut().get(dtype, dtype)
        return torch.round(data, out=out).to(dtype)

    @_torchmethod(func_type='data_func')
    def clip(data, a_min=None, a_max=None, out=None):
        return torch.clamp(data, a_min, a_max, out=out)

    @_torchmethod(func_type='data_func')
    def softmax(data, axis=None):
        return torch.softmax(data, dim=axis)


class NumpyImpls(object):
    """
    Numpy backend for the ArrayAPI API
    """

    is_tensor = False
    is_numpy = True

    hstack = _numpymethod(func_type='array_sequence')(np.hstack)
    vstack = _numpymethod(func_type='array_sequence')(np.vstack)

    @_numpymethod(func_type='array_sequence')
    def cat(datas, axis=-1):
        return np.concatenate(datas, axis=axis)

    @_numpymethod(func_type='data_func')
    def atleast_nd(arr, n, front=False):
        import kwarray
        return kwarray.atleast_nd(arr, n, front)

    @_numpymethod(func_type='data_func')
    def view(data, *shape):
        data_ = data.reshape(*shape)
        return data_

    @_numpymethod(func_type='data_func')
    def take(data, indices, axis=None):
        return data.take(indices, axis=axis)

    @_numpymethod(func_type='data_func')
    def compress(data, flags, axis=None):
        return data.compress(flags, axis=axis)

    @_numpymethod(func_type='data_func')
    def repeat(data, repeats, axis=None):
        if not (axis is None or isinstance(repeats, int)):
            raise NotImplementedError('torch version of non-int repeats is not implemented')
        return np.repeat(data, repeats, axis=axis)

    @_numpymethod(func_type='data_func')
    def tile(data, reps):
        return np.tile(data, reps)

    @_numpymethod(func_type='data_func')
    def T(data):
        return data.T

    @_numpymethod(func_type='data_func')
    def transpose(data, axes):
        return np.transpose(data, axes)

    @_numpymethod(func_type='data_func')
    def numel(data):
        return data.size

    # --- Allocators ----

    @_numpymethod
    def empty_like(data, dtype=None):
        return np.empty_like(data, dtype=dtype)

    @_numpymethod
    def full_like(data, fill_value, dtype=None):
        return np.full_like(data, fill_value, dtype=dtype)

    @_numpymethod
    def zeros_like(data, dtype=None):
        return np.zeros_like(data, dtype=dtype)

    @_numpymethod
    def ones_like(data, dtype=None):
        return np.ones_like(data, dtype=dtype)

    @_numpymethod(func_type='shape_creation')
    def full(shape, fill_value, dtype=float):
        return np.full(shape, fill_value, dtype=dtype)

    @_numpymethod(func_type='shape_creation')
    def empty(shape, dtype=float):
        return np.full(shape, dtype=dtype)

    @_numpymethod(func_type='shape_creation')
    def zeros(shape, dtype=float):
        return np.zeros(shape, dtype=dtype)

    @_numpymethod(func_type='shape_creation')
    def ones(shape, dtype=float):
        return np.ones(shape, dtype=dtype)

    # -------

    @_numpymethod
    def argmax(data, axis=None):
        return np.argmax(data, axis=axis)

    @_numpymethod
    def argsort(data, axis=-1, descending=False):
        sortx = np.argsort(data, axis=axis)
        if descending:
            sortx = np.flip(sortx, axis=axis)
        return sortx

    @_numpymethod
    def max(data, axis=None):
        return data.max(axis=axis)

    @_numpymethod
    def max_argmax(data, axis=None):
        return data.max(axis=axis), np.argmax(data, axis=axis)

    @_numpymethod
    def sum(data, axis=None):
        return data.sum(axis=axis)

    @_numpymethod
    def maximum(data1, data2, out=None):
        return np.maximum(data1, data2, out=out)

    @_numpymethod
    def minimum(data1, data2, out=None):
        return np.minimum(data1, data2, out=out)

    # @_numpymethod
    # def matmul(data1, data2, out=None):
    #     return np.matmul(data1, data2, out=out)
    matmul = _numpymethod(np.matmul)

    nan_to_num = _numpymethod(np.nan_to_num)

    log = _numpymethod(np.log)
    log2 = _numpymethod(np.log2)

    any = _numpymethod(np.any)
    all = _numpymethod(np.all)

    copy = _numpymethod(np.copy)

    nonzero = _numpymethod(np.nonzero)

    @_numpymethod(func_type='data_func')
    def astype(data, dtype, copy=True):
        return data.astype(dtype, copy=copy)

    @_numpymethod(func_type='data_func')
    def tensor(data, device=ub.NoParam):
        if torch is None:
            raise Exception('torch is not available')
        data = torch.from_numpy(np.ascontiguousarray(data))
        if device is not ub.NoParam:
            data = data.to(device)
        return data

    @_numpymethod(func_type='data_func')
    def numpy(data):
        return data

    @_numpymethod(func_type='data_func')
    def tolist(data):
        return data.tolist()

    @_numpymethod(func_type='data_func')
    def contiguous(data):
        return np.ascontiguousarray(data)

    @_numpymethod(func_type='data_func')
    def pad(data, pad_width, mode='constant'):
        return np.pad(data, pad_width, mode=mode)

    @_numpymethod(func_type='data_func')
    def asarray(data, dtype=None):
        """
        Cast data into a numpy representation
        """
        if torch is not None and isinstance(data, torch.Tensor):
            data = data.cpu().numpy()
        return np.asarray(data, dtype=dtype)

    ensure = asarray

    @_numpymethod(func_type='data_func')
    def dtype_kind(data):
        return data.dtype.kind

    @_numpymethod(func_type='data_func')
    def floor(data, out=None):
        return np.floor(data, out=out)

    @_numpymethod(func_type='data_func')
    def ceil(data, out=None):
        return np.ceil(data, out=out)

    @_numpymethod(func_type='data_func')
    def ifloor(data, out=None):
        return np.floor(data, out=out).astype(np.int32)

    @_numpymethod(func_type='data_func')
    def iceil(data, out=None):
        return np.ceil(data, out=out).astype(np.int32)

    @_numpymethod(func_type='data_func')
    def round(data, decimals=0, out=None):
        return np.round(data, decimals=decimals, out=out)

    @_numpymethod(func_type='data_func')
    def iround(data, out=None, dtype=int):
        return np.round(data, out=out).astype(dtype)

    clip = _numpymethod(np.clip)

    @_numpymethod(func_type='data_func')
    def softmax(data, axis=None):
        from scipy import special
        return special.softmax(data, axis=axis)


class ArrayAPI(object):
    """
    Compatability API between torch and numpy.

    The API defines classmethods that work on both Tensors and ndarrays.  As
    such the user can simply use ``kwarray.ArrayAPI.<funcname>`` and it will
    return the expected result for both Tensor and ndarray types.

    However, this is inefficient because it requires us to check the type of
    the input for every API call. Therefore it is recommended that you use the
    :func:`ArrayAPI.coerce` function, which takes as input the data you want to
    operate on. It performs the type check once, and then returns another
    object that defines with an identical API, but specific to the given data
    type. This means that we can ignore type checks on future calls of the
    specific implementation. See examples for more details.


    Example:
        >>> # Use the easy-to-use, but inefficient array api
        >>> # xdoctest: +REQUIRES(module:torch)
        >>> take = ArrayAPI.take
        >>> np_data = np.arange(0, 143).reshape(11, 13)
        >>> pt_data = torch.LongTensor(np_data)
        >>> indices = [1, 3, 5, 7, 11, 13, 17, 21]
        >>> idxs0 = [1, 3, 5, 7]
        >>> idxs1 = [1, 3, 5, 7, 11]
        >>> assert np.allclose(take(np_data, indices), take(pt_data, indices))
        >>> assert np.allclose(take(np_data, idxs0, 0), take(pt_data, idxs0, 0))
        >>> assert np.allclose(take(np_data, idxs1, 1), take(pt_data, idxs1, 1))

    Example:
        >>> # Use the easy-to-use, but inefficient array api
        >>> # xdoctest: +REQUIRES(module:torch)
        >>> compress = ArrayAPI.compress
        >>> np_data = np.arange(0, 143).reshape(11, 13)
        >>> pt_data = torch.LongTensor(np_data)
        >>> flags = (np_data % 2 == 0).ravel()
        >>> f0 = (np_data % 2 == 0)[:, 0]
        >>> f1 = (np_data % 2 == 0)[0, :]
        >>> assert np.allclose(compress(np_data, flags), compress(pt_data, flags))
        >>> assert np.allclose(compress(np_data, f0, 0), compress(pt_data, f0, 0))
        >>> assert np.allclose(compress(np_data, f1, 1), compress(pt_data, f1, 1))

    Example:
        >>> # Use ArrayAPI to coerce an identical API that doesnt do type checks
        >>> # xdoctest: +REQUIRES(module:torch)
        >>> import kwarray
        >>> np_data = np.arange(0, 15).reshape(3, 5)
        >>> pt_data = torch.LongTensor(np_data)
        >>> # The new ``impl`` object has the same API as ArrayAPI, but works
        >>> # specifically on torch Tensors.
        >>> impl = kwarray.ArrayAPI.coerce(pt_data)
        >>> flat_data = impl.view(pt_data, -1)
        >>> print('flat_data = {!r}'.format(flat_data))
        flat_data = tensor([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14])
        >>> # The new ``impl`` object has the same API as ArrayAPI, but works
        >>> # specifically on numpy ndarrays.
        >>> impl = kwarray.ArrayAPI.coerce(np_data)
        >>> flat_data = impl.view(np_data, -1)
        >>> print('flat_data = {!r}'.format(flat_data))
        flat_data = array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14])
    """

    _torch = TorchImpls
    _numpy = NumpyImpls

    @staticmethod
    def impl(data):
        """
        Returns a namespace suitable for operating on the input data type

        Args:
            data (ndarray | Tensor): data to be operated on

        """
        if torch is not None and torch.is_tensor(data):
            return TorchImpls
        else:
            return NumpyImpls

    @staticmethod
    def coerce(data):
        """
        Coerces some form of inputs into an array api (either numpy or torch).
        """
        if isinstance(data, six.string_types):
            from kwarray import arrayapi
            if data in ['torch', 'tensor']:
                return arrayapi.TorchImpls
            elif data == 'numpy':
                return arrayapi.NumpyImpls
            else:
                raise KeyError(data)
        else:
            return ArrayAPI.impl(data)

    @_apimethod(func_type='array_sequence')
    def cat(datas, *args, **kwargs):
        impl = ArrayAPI.impl(datas[0])
        return impl.cat(datas, *args, **kwargs)

    @_apimethod(func_type='array_sequence')
    def hstack(datas, *args, **kwargs):
        impl = ArrayAPI.impl(datas[0])
        return impl.hstack(datas, *args, **kwargs)

    @_apimethod(func_type='array_sequence')
    def vstack(datas, *args, **kwargs):
        impl = ArrayAPI.impl(datas[0])
        return impl.vstack(datas, *args, **kwargs)

    take = _apimethod('take')
    compress = _apimethod('compress')

    repeat = _apimethod('repeat')
    tile = _apimethod('tile')

    view = _apimethod('view')
    numel = _apimethod('numel')
    atleast_nd = _apimethod('atleast_nd')

    full_like = _apimethod('full_like')
    ones_like = _apimethod('ones_like')
    zeros_like = _apimethod('zeros_like')
    empty_like = _apimethod('empty_like')

    sum = _apimethod('sum')
    argmax = _apimethod('argmax')
    argsort = _apimethod('argsort')
    max = _apimethod('max')
    maximum = _apimethod('maximum')
    minimum = _apimethod('minimum')

    matmul = _apimethod('matmul')

    astype = _apimethod('astype')
    nonzero = _apimethod('nonzero')

    nan_to_num = _apimethod('nan_to_num')

    tensor = _apimethod('tensor')
    numpy = _apimethod('numpy')
    tolist = _apimethod('tolist')
    asarray = _apimethod('asarray')
    asarray = _apimethod('ensure')

    T = _apimethod('T')
    transpose = _apimethod('transpose')

    contiguous = _apimethod('contiguous')
    pad = _apimethod('pad')

    dtype_kind = _apimethod('dtype_kind')

    max_argmax = _apimethod('max_argmax')

    # ones = _apimethod('ones', func_type='shape_creation')
    # full = _apimethod('full', func_type='shape_creation')
    # empty = _apimethod('empty', func_type='shape_creation')
    # zeros = _apimethod('zeros', func_type='shape_creation')

    any = _apimethod('any', func_type='data_func')
    all = _apimethod('all', func_type='data_func')

    log2 = _apimethod('log2', func_type='data_func')
    log = _apimethod('log', func_type='data_func')
    copy = _apimethod('copy', func_type='data_func')

    iceil = _apimethod('iceil', func_type='data_func')
    ifloor = _apimethod('ifloor', func_type='data_func')
    floor = _apimethod('floor', func_type='data_func')
    ceil = _apimethod('ceil', func_type='data_func')

    round = _apimethod('round', func_type='data_func')
    iround = _apimethod('iround', func_type='data_func')

    clip = _apimethod('clip', func_type='data_func')

    softmax = _apimethod('softmax', func_type='data_func')


TorchNumpyCompat = ArrayAPI  # backwards compat


if __debug__ and torch is not None:
    _REGISTERY._ensure_datamethods_names_are_registered()


@ub.memoize
def _torch_dtype_lut():
    lut = {}

    # Handle nonstandard alias dtype names
    lut['double'] = torch.double
    lut['long'] = torch.long

    # Handle floats
    for k in [np.float16, 'float16']:
        lut[k] = torch.float16
    for k in [np.float32, 'float32']:
        lut[k] = torch.float32
    for k in [np.float64, 'float64']:
        lut[k] = torch.float64

    if torch.float == torch.float32:
        lut['float'] = torch.float32
    else:
        raise AssertionError('dont think this can happen')

    if np.float_ == np.float32:
        lut[float] = torch.float32
    elif np.float_ == np.float64:
        lut[float] = torch.float64
    else:
        raise AssertionError('dont think this can happen')

    # Handle signed integers
    for k in [np.int8, 'int8']:
        lut[k] = torch.int8
    for k in [np.int16, 'int16']:
        lut[k] = torch.int16
    for k in [np.int32, 'int32']:
        lut[k] = torch.int32
    for k in [np.int64, 'int64']:
        lut[k] = torch.int64

    if np.int_ == np.int32:
        lut[int] = torch.int32
    elif np.int_ == np.int64:
        lut[int] = torch.int64
    else:
        raise AssertionError('dont think this can happen')

    if torch.int == torch.int32:
        lut['int'] = torch.int32
    else:
        raise AssertionError('dont think this can happen')

    # Handle unsigned integers
    for k in [np.uint8, 'uint8']:
        lut[k] = torch.uint8

    # import torch.utils.data
    # cant use torch.utils.data.dataloader.numpy_type_map directly because it
    # maps to tensor types not dtypes, but we can use it to check
    check = False
    if check:
        for k, v in torch.utils.data.dataloader.numpy_type_map.items():
            assert lut[k] == v.dtype
    return lut


def dtype_info(dtype):
    """
    Args:
        dtype (type): a numpy, torch, or python numeric data type

    Returns:
        struct: an iinfo of finfo structure depending on the input type.

    References:
        https://higra.readthedocs.io/en/stable/_modules/higra/hg_utils.html#dtype_info

    Example:
        >>> results = []
        >>> results += [dtype_info(float)]
        >>> results += [dtype_info(int)]
        >>> results += [dtype_info(complex)]
        >>> results += [dtype_info(np.float32)]
        >>> results += [dtype_info(np.int32)]
        >>> results += [dtype_info(np.uint32)]
        >>> if hasattr(np, 'complex256'):
        >>>     results += [dtype_info(np.complex256)]
        >>> if torch is not None:
        >>>     results += [dtype_info(torch.float32)]
        >>>     results += [dtype_info(torch.int64)]
        >>>     results += [dtype_info(torch.complex64)]
        >>> for info in results:
        >>>     print('info = {!r}'.format(info))
        >>> for info in results:
        >>>     print('info.bits = {!r}'.format(info.bits))
    """
    if torch is not None and isinstance(dtype, torch.dtype):
        # Using getattr on is_complex for torch 1.4
        if dtype.is_floating_point or getattr(dtype, 'is_complex', False):
            try:
                info = torch.finfo(dtype)
            except Exception:
                # if not hasattr(dtype, 'is_complex'):
                """
                # Helper for dev
                zs = [torch.complex32, torch.complex64, torch.complex128]
                for z in zs:
                    z = torch.finfo(z)
                    args = [f + '=' + repr(getattr(z, f))
                        for f in finfo_ducktype._fields]
                    print('finfo_ducktype({})'.format(', '.join(args)))
                """
                # Return a ducktype for older torches without complex finfo
                # https://github.com/pytorch/pytorch/issues/35954
                from collections import namedtuple
                finfo_ducktype = namedtuple('finfo_ducktype', [
                    'bits', 'dtype', 'eps', 'max', 'min', 'resolution',
                    'tiny'])
                from collections import namedtuple
                finfo_ducktype(bits=32, dtype='float32',
                               eps=1.1920928955078125e-07,
                               max=3.4028234663852886e+38,
                               min=-3.4028234663852886e+38,
                               resolution=1e-06,
                               tiny=1.1754943508222875e-38)
                if dtype == torch.complex32:
                    info = finfo_ducktype(bits=16, dtype='float16',
                                          eps=0.0009765625, max=65504.0,
                                          min=-65504.0, resolution=0.001,
                                          tiny=6.103515625e-05)
                elif dtype == torch.complex64:
                    info = finfo_ducktype(bits=32, dtype='float32',
                                          eps=1.1920928955078125e-07,
                                          max=3.4028234663852886e+38,
                                          min=-3.4028234663852886e+38,
                                          resolution=1e-06,
                                          tiny=1.1754943508222875e-38)
                elif dtype == torch.complex128:
                    info = finfo_ducktype(bits=64, dtype='float64',
                                          eps=2.220446049250313e-16,
                                          max=1.7976931348623157e+308,
                                          min=-1.7976931348623157e+308,
                                          resolution=1e-15,
                                          tiny=2.2250738585072014e-308)
                else:
                    raise TypeError(dtype)
        else:
            info = torch.iinfo(dtype)
    else:
        np_dtype = np.dtype(dtype)
        if np_dtype.kind in {'f', 'c'}:
            info = np.finfo(np_dtype)
        else:
            info = np.iinfo(np_dtype)
    return info
